from .helpers import Order
from .helpers import OrderType
from .helpers import OrderPriority
from .helpers import Priority
from .helpers import _unpack_list


class Commands:
    invalid = 4294967295

    def __init__(self, api, ffi):
        self._api = api
        self._ffi = ffi

    def orders(self, unit: int) -> list[Order]:
        os = self._ffi.new("struct UwOrders *")
        self._api.uwOrders(unit, os)
        return [Order.from_c(o) for o in _unpack_list(self._ffi, os, "orders")]

    def order(self, unit: int, order: Order):
        o = self._ffi.new("struct UwOrder *")
        o.entity = order.entity
        o.position = order.position
        o.order = int(order.order_type)
        o.priority = int(order.priority)
        self._api.uwOrder(unit, o)

    def stop(self) -> Order:
        return Order(entity=self.invalid, position=self.invalid, order_type=OrderType.Stop, priority=OrderPriority.User)

    def guard(self) -> Order:
        return Order(entity=self.invalid, position=self.invalid, order_type=OrderType.Guard,
                     priority=OrderPriority.User)

    def run_to_position(self, position: int) -> Order:
        return Order(entity=self.invalid, position=position, order_type=OrderType.Run,
                     priority=OrderPriority.User)

    def run_to_entity(self, entity: int) -> Order:
        return Order(entity=entity, position=self.invalid, order_type=OrderType.Run,
                     priority=OrderPriority.User)

    def fight_to_position(self, position: int) -> Order:
        return Order(entity=self.invalid, position=position, order_type=OrderType.Fight,
                     priority=OrderPriority.User)

    def fight_to_entity(self, entity: int) -> Order:
        return Order(entity=entity, position=self.invalid, order_type=OrderType.Fight,
                     priority=OrderPriority.User)

    def command_self_destruct(self, unit: int):
        self._api.uwCommandSelfDestruct(unit)

    def command_place_construction(self, proto: int, position: int, yaw: float = 0):
        self._api.uwCommandPlaceConstruction(proto, position, yaw)

    def command_set_recipe(self, unit: int, recipe: int):
        self._api.uwCommandSetRecipe(unit, recipe)

    def command_set_priority(self, unit: int, priority: Priority):
        self._api.uwCommandSetPriority(unit, priority)

    def command_load(self, unit: int, resource_type: int):
        self._api.uwCommandLoad(unit, resource_type)

    def command_unload(self, unit: int):
        self._api.uwCommandUnload(unit)

    def command_move(self, unit: int, position: int, yaw: float = 0):
        self._api.uwCommandMove(unit, position, yaw)

    def command_aim(self, unit: int, target: int):
        self._api.uwCommandAim(unit, target)

    def command_renounce_control(self, unit: int):
        self._api.uwCommandRenounceControl(unit)



from enum import Enum
from enum import Flag
from enum import IntEnum
from enum import IntFlag
from typing import List
from typing import Any


def _c_str(s: str) -> bytes:
    return bytes(s, encoding="utf-8")


def _to_str(ffi, s) -> str:
    return ffi.string(s).decode("utf-8")


def _unpack_list(ffi, data, attribute="ids") -> List[Any]:
    if data.count > 0:
        return ffi.unpack(getattr(data, attribute), data.count)
    return []


class Severity(Enum):
    Note = 0
    Hint = 1
    Warning = 2
    Info = 3
    Error = 4
    Critical = 5


class ConnectionState(Enum):
    NONE = 0
    Connecting = 1
    Connected = 2
    Disconnecting = 3
    Error = 4


class MapState(Enum):
    NONE = 0
    Downloading = 1
    Loading = 2
    Loaded = 3
    Unloading = 4
    Error = 5


class GameState(Enum):
    NONE = 0
    Session = 1
    Preparation = 2
    Game = 3
    Finish = 4


class OrderType(IntEnum):
    NONE = 0
    Stop = 1
    Guard = 2
    Run = 3
    Fight = 4
    Load = 5
    Unload = 6
    SelfDestruct = 7


class OrderPriority(IntFlag):
    NONE = 0
    Assistant = 1 << 0
    User = 1 << 1
    Enqueue = 1 << 2
    Repeat = 1 << 3


class Prototype(Enum):
    NONE = 0
    Resource = 1
    Recipe = 2
    Construction = 3
    Unit = 4


class OverviewFlags(Flag):
    NONE = 0
    Resource = 1 << 0
    Construction = 1 << 1
    MobileUnit = 1 << 2
    StaticUnit = 1 << 3
    Unit = MobileUnit | StaticUnit


class Priority(Enum):
    Disabled = 0,
    Normal = 1,
    High = 2,


class LogCallback:
    def __init__(self, message: str, component: str, severity: Severity):
        self.message = message
        self.component = component
        self.severity = severity

    @staticmethod
    def from_c(ffi, data):
        return LogCallback(_to_str(ffi, data.message), _to_str(ffi, data.component), Severity(data.severity))


class Order:
    def __init__(self, entity: int, position: int, order_type: OrderType, priority: OrderPriority):
        self.entity = entity
        self.position = position
        self.order_type = order_type
        self.priority = priority

    @staticmethod
    def from_c(data):
        return Order(data.entity, data.position, OrderType(data.order), OrderPriority(data.priority))


class ShootingUnit:
    def __init__(self, data):
        self.position = data.position
        self.force = data.force
        self.prototype = data.prototype
        self.id = data.id

    @staticmethod
    def _from_c(self, data):
        self.position = data.position
        self.force = data.force
        self.prototype = data.prototype
        self.id = data.id


class ShootingData:
    def __init__(self, shooter: ShootingUnit, target: ShootingUnit):
        self.shooter = shooter
        self.target = target

    @staticmethod
    def from_c(data):
        return ShootingData(ShootingUnit(data.shooter), ShootingUnit(data.target))



import json

from typing import Any
from typing import Optional

from .helpers import MapState
from .helpers import Prototype
from .helpers import _c_str
from .helpers import _to_str
from .helpers import _unpack_list


class ProtoGeneric:
    def __init__(self, _type: Prototype, name: str, _json: str):
        self.type = _type
        self.name = name
        self.json = _json


class Prototypes:
    def __init__(self, api, ffi, game):
        self._api = api
        self._ffi = ffi
        self.game = game
        self.game.add_map_state_callback(self._map_state_changed)

        self._hit_chances_table: dict[str, Any]
        self._terrain_types_table: dict[str, Any]

        self._all: list[int] = []
        self._types: dict[int, ProtoGeneric] = {}
        self._resources: dict[int, Any] = {}
        self._recipes: dict[int, Any] = {}
        self._constructions: dict[int, Any] = {}
        self._units: dict[int, Any] = {}

    def all(self) -> list[int]:
        return self._all

    def type(self, _id: int) -> Prototype:
        return self._types[_id].type if _id in self._types else Prototype.NONE

    def name(self, _id: int) -> str:
        return self._types[_id].name if _id in self._types else ""

    def json(self, _id: int) -> str:
        return self._types[_id].json if _id in self._types else ""

    def resource(self, _id: int) -> Optional[dict]:
        return self._resources.get(_id)

    def recipes(self, _id: int) -> Optional[dict]:
        return self._recipes.get(_id)

    def construction(self, _id: int) -> Optional[dict]:
        return self._constructions.get(_id)

    def unit(self, _id: int) -> Optional[dict]:
        return self._units.get(_id)

    def hit_chances_table(self):
        return self._hit_chances_table

    def terrain_types_table(self):
        return self._terrain_types_table

    def _all_ids(self) -> list[int]:
        ids = self._ffi.new("struct UwIds *")
        self._api.uwAllPrototypes(ids)
        return _unpack_list(self._ffi, ids)

    def _load_prototypes(self):
        print("loading prototypes")

        self._all = []
        self._types = {}
        self._resources = {}
        self._recipes = {}
        self._constructions = {}
        self._units = {}

        for i in self._all_ids():
            _type = Prototype(self._api.uwPrototypeType(i))
            js = json.loads(_to_str(self._ffi, self._api.uwPrototypeJson(i)))
            if _type == Prototype.Resource:
                self._resources[i] = js
            elif _type == Prototype.Recipe:
                self._recipes[i] = js
            elif _type == Prototype.Construction:
                self._constructions[i] = js
            elif _type == Prototype.Unit:
                self._units[i] = js

            self._types[i] = ProtoGeneric(_type, js["name"], js)
            self._all.append(i)

        print("prototypes loaded")

    def _load_definitions(self):
        print("loading definitions")

        defs = json.loads(_to_str(self._ffi, self._api.uwDefinitionsJson()))
        self._hit_chances_table = defs["hitChancesTable"]
        self._terrain_types_table = defs["terrainTypesTable"]

        print("definitions loaded")

    def _map_state_changed(self, state: MapState):
        if state == MapState.Loaded:
            self._load_prototypes()
            self._load_definitions()



import math

from .helpers import MapState
from .helpers import OverviewFlags
from .helpers import _unpack_list


class Vector3:
    def __init__(self, x: float, y: float, z: float):
        self.x = x
        self.y = y
        self.z = z


class Map:
    def __init__(self, api, ffi, game):
        self._api = api
        self._ffi = ffi
        self._game = game

        self._game.add_map_state_callback(self._map_state_changed)
        self._game.add_update_callback(self._updating)

        self._name: str = ""
        self._guid: str = ""
        self._path: str = ""
        self._max_players: int = 0
        self._positions: list[Vector3] = []
        self._ups: list[Vector3] = []
        self._neighbors: list[list[int]] = []
        self._terrains: list[bytes] = []
        self._overview: list[OverviewFlags] = []

    def name(self) -> str:
        return self._name

    def guid(self) -> str:
        return self._guid

    def path(self) -> str:
        return self._path

    def max_players(self) -> int:
        return self._max_players

    def positions(self) -> list[Vector3]:
        return self._positions

    def ups(self) -> list[Vector3]:
        return self._ups

    def neighbors(self) -> list[list[int]]:
        return self._neighbors

    def neighbors_on_pos(self, pos: int) -> list[int]:
        return self._neighbors[pos]

    def terrains(self) -> list[bytes]:
        return self._terrains

    def overview(self) -> list[OverviewFlags]:
        return self._overview

    def entities(self, position: int) -> list[int]:
        ns = self._ffi.new("struct UwIds *")
        self._api.uwOverviewIds(position, ns)
        return _unpack_list(self._ffi, ns)

    def area_range(self, point: Vector3, radius: float) -> list[int]:
        tiles = self._ffi.new("struct UwIds *")
        self._api.uwAreaRange(point.x, point.y, point.z, radius, tiles)
        return _unpack_list(self._ffi, tiles)

    def area_connected(self, position: int, radius: float) -> list[int]:
        tiles = self._ffi.new("struct UwIds *")
        self._api.uwAreaConnected(position, radius, tiles)
        return _unpack_list(self._ffi, tiles)

    def area_neighborhood(self, position: int, radius: float) -> list[int]:
        tiles = self._ffi.new("struct UwIds *")
        self._api.uwAreaNeighborhood(position, radius, tiles)
        return _unpack_list(self._ffi, tiles)

    def area_extended(self, position: int, radius: float) -> list[int]:
        tiles = self._ffi.new("struct UwIds *")
        self._api.uwAreaExtended(position, radius, tiles)
        return _unpack_list(self._ffi, tiles)

    def test_visible(self, a: Vector3, b: Vector3) -> bool:
        return self._ffi.uwTestVisible(a.x, a.y, a.z, b.x, b.y, b.z)

    def test_shooting(self, shooter_position: int, shooter_proto: int, target_position: int, target_proto: int):
        return self._ffi.uwTestShooting(shooter_position, shooter_proto, target_position, target_proto)

    def distance_line(self, ai: int, bi: int) -> float:
        a: Vector3 = self._positions[ai]
        b: Vector3 = self._positions[bi]
        x: float = a.x - b.x
        y: float = a.y - b.y
        z: float = a.z - b.z
        x *= x
        y *= y
        z *= z
        return math.sqrt(x + y + z)

    def distance_estimate(self, a: int, b: int) -> float:
        return self._api.uwDistanceEstimate(a, b)

    def yaw(self, a: int, b: int) -> float:
        return self._api.uwDistanceEstimate(a, b)

    def test_construction_placement(self, construction_prototype: int, position: int) -> bool:
        return self._api.uwTestConstructionPlacement(construction_prototype, position)

    def find_construction_placement(self, construction_prototype: int, position: int) -> bool:
        return self._api.uwFindConstructionPlacement(construction_prototype, position)

    def _load(self):
        print("loading map")
        self._positions = []
        self._ups = []
        self._neighbors = []
        self._terrains = []
        self._overview = []

        info = self._ffi.new("struct UwMapInfo *")
        self._api.uwMapInfo(info)
        self._name = self._ffi.string(info.name)
        self._guid = self._ffi.string(info.guid)
        self._path = self._ffi.string(info.path)
        self._max_players = info.maxPlayers
        print(f"map name: {self._name}")
        print(f"map guid: {self._guid}")

        count = self._api.uwTilesCount()
        tile = self._ffi.new("struct UwTile *")
        for i in range(count):
            self._api.uwTile(i, tile)
            p = Vector3(tile.position[0], tile.position[1], tile.position[2])
            self._positions.append(p)
            u = Vector3(tile.up[0], tile.up[1], tile.up[2])
            self._ups.append(u)
            n = self._ffi.unpack(tile.neighborsIndices, tile.neighborsCount) if tile.neighborsCount > 0 else []
            if n:
                self._neighbors.append(n)
            self._terrains.append(tile.terrain)

        print("map loaded")

    def _map_state_changed(self, map_state: MapState):
        if map_state == MapState.Loaded:
            self._load()

    def _updating(self, stepping: bool):
        if stepping:
            ex = self._ffi.new("struct UwOverviewExtract *")
            self._api.uwOverviewExtract(ex)
            if ex.count > 0:
                self._overview = [OverviewFlags(i) for i in _unpack_list(self._ffi, ex, "flags")]
        else:
            self._overview = []



from typing import Any
from enum import Enum

from .helpers import _unpack_list


class Policy(Enum):
    NONE = 0
    Self = 1
    Ally = 2
    Neutral = 3
    Enemy = 4


class Entity:
    def __init__(self, world):
        self._world = world

    def has(self, component: str):
        return hasattr(self, component)

    def own(self) -> bool:
        return self.has("Owner") and self.Owner.force == self._world.my_force()

    def policy(self) -> Policy:
        if not self.has("Owner"):
            return Policy.NONE

        return self._world.policy(self.Owner.force)


class World:
    def __init__(self, api, ffi, game):
        self._api = api
        self._ffi = ffi
        self._game = game

        self._my_force: int = 0
        self._entities: dict[int, Any] = {}
        self._policies: dict[int, Policy] = {}

        self._game.add_update_callback(self._updating)

    def my_force(self) -> int:
        return self._my_force

    def entities(self) -> dict[int, Any]:
        return self._entities

    def entity(self, _id: int) -> Any:
        return self._entities[_id]

    def policy(self, force: int) -> Policy:
        return self._policies.get(force, Policy.NONE)

    def _all_ids(self) -> list[int]:
        ids = self._ffi.new("struct UwIds *")
        self._api.uwAllEntities(ids)
        return _unpack_list(self._ffi, ids)

    def _modified_ids(self) -> list[int]:
        ids = self._ffi.new("struct UwIds *")
        self._api.uwModifiedEntities(ids)
        return _unpack_list(self._ffi, ids)

    def _update_removed(self):
        all_ids = set(self._all_ids())
        removed = []
        for _id in self._entities.keys():
            if _id not in all_ids:
                removed.append(_id)
        for _id in removed:
            del self._entities[_id]

    def _maybe_assign_or_remove(self, e, o, fetch_method):
        struct = fetch_method.replace("uwFetch", "Uw")
        field = fetch_method.replace("uwFetch", "").replace("Component", "")
        tmp = self._ffi.new(f"struct {struct} *")
        if getattr(self._api, fetch_method)(e, tmp):
            # print(field)
            setattr(o, field, tmp)
            # print(o)
            # print(dir(o))
        else:
            if hasattr(o, field):
                delattr(o, field)

    def _update_modified(self):
        for _id in self._modified_ids():
            o = self._entities.get(_id, Entity(self))
            o.Id = _id
            self._entities[_id] = o
            e = self._api.uwEntityPointer(_id)

            self._maybe_assign_or_remove(e, o, "uwFetchProtoComponent")
            self._maybe_assign_or_remove(e, o, "uwFetchOwnerComponent")
            self._maybe_assign_or_remove(e, o, "uwFetchControllerComponent")
            self._maybe_assign_or_remove(e, o, "uwFetchPositionComponent")
            self._maybe_assign_or_remove(e, o, "uwFetchUnitComponent")
            self._maybe_assign_or_remove(e, o, "uwFetchLifeComponent")
            self._maybe_assign_or_remove(e, o, "uwFetchMoveComponent")
            self._maybe_assign_or_remove(e, o, "uwFetchAimComponent")
            self._maybe_assign_or_remove(e, o, "uwFetchRecipeComponent")
            self._maybe_assign_or_remove(e, o, "uwFetchUpdateTimestampComponent")
            self._maybe_assign_or_remove(e, o, "uwFetchRecipeStatisticsComponent")
            self._maybe_assign_or_remove(e, o, "uwFetchPriorityComponent")
            self._maybe_assign_or_remove(e, o, "uwFetchAmountComponent")
            self._maybe_assign_or_remove(e, o, "uwFetchAttachmentComponent")
            self._maybe_assign_or_remove(e, o, "uwFetchPlayerComponent")
            self._maybe_assign_or_remove(e, o, "uwFetchForceComponent")
            self._maybe_assign_or_remove(e, o, "uwFetchForceDetailsComponent")
            self._maybe_assign_or_remove(e, o, "uwFetchForeignPolicyComponent")
            self._maybe_assign_or_remove(e, o, "uwFetchDiplomacyProposalComponent")

    def _update_policies(self):
        self._policies = {}
        for e in self._entities.values():
            if not hasattr(e, "ForeignPolicy"):
                continue
            fp = getattr(e, "ForeignPolicy")
            forces = self._ffi.unpack(fp.forces, 2)
            policy = Policy(fp.policy)
            if forces[0] == self._my_force:
                self._policies[forces[1]] = policy
            if forces[1] == self._my_force:
                self._policies[forces[0]] = policy

    def _updating(self, stepping: bool):
        player = self._ffi.new("struct UwMyPlayer *")
        self._api.uwMyPlayer(player)
        self._my_force = player.forceEntityId

        self._update_removed()
        self._update_modified()
        self._update_policies()



import os
import sys

from cffi import FFI
from typing import Callable

from .commands import Commands
from .helpers import _c_str
from .helpers import _to_str
from .helpers import Severity
from .helpers import ConnectionState
from .helpers import MapState
from .helpers import LogCallback
from .helpers import GameState
from .helpers import ShootingData
from .prototypes import Prototypes
from .map import Map
from .world import World
from .helpers import _unpack_list

_LIB_NAME_PATTERN = "libunnatural-uwapi{}.{}"
_STEAM_PATH_ENV_NAME = 'UNNATURAL_ROOT'


def get_lib_name(hardened=True):
    return _LIB_NAME_PATTERN.format("-hard" if hardened else "", "dll" if sys.platform == "win32" else "so")


def get_default_steam_location():
    if sys.platform == "win32":
        return "C:\\Program Files (x86)\\Steam\\common\\Unnatural Worlds\\bin"
    return os.path.expanduser("~/.local/share/Steam/common/Unnatural Worlds/bin")


def get_steam_path(steam_path: str) -> str:
    if steam_path != "":
        return steam_path
    steam_path = os.environ.get(_STEAM_PATH_ENV_NAME, "")
    if steam_path != "":
        return steam_path
    return get_default_steam_location()


class Game:
    def __init__(self, steam_path: str = "", hardened: bool = True):
        steam_path = get_steam_path(steam_path)

        api_def = open(os.path.join(os.path.split(os.path.abspath(__file__))[0], "bots.h"), "r").read()
        os.chdir(steam_path)

        self._ffi = FFI()
        self._ffi.cdef(api_def)
        self._api = self._ffi.dlopen(os.path.join(steam_path, get_lib_name(hardened=hardened)))
        # print(f"api version {self._api.UW_VERSION}")
        self._api.uwInitialize(self._api.UW_VERSION)

        self._connection_state_changed_handler = []
        self._updating_handler = []
        self._game_state_changed_handler = []
        self._map_state_changed_handler = []
        self._shooting_handler = []

        self._exception_delegate = self._ffi.callback("UwExceptionCallbackType", self._exception_callback)
        self._exception_callback_func = self._api.uwSetExceptionCallback(self._exception_delegate)

        self._log_delegate = self._ffi.callback("UwLogCallbackType", self._log_callback)
        self._log_callback_func = self._api.uwSetLogCallback(self._log_delegate)

        self._connection_state_delegate = self._ffi.callback("UwConnectionStateCallbackType",
                                                             self._connection_state_callback)
        self._connection_state_callback_func = self._api.uwSetConnectionStateCallback(self._connection_state_delegate)

        self._game_state_delegate = self._ffi.callback("UwGameStateCallbackType", self._game_state_callback)
        self._game_state_callback_func = self._api.uwSetGameStateCallback(self._game_state_delegate)

        self._map_state_delegate = self._ffi.callback("UwMapStateCallbackType", self._map_state_callback)
        self._map_state_callback_func = self._api.uwSetMapStateCallback(self._map_state_delegate)

        self._update_delegate = self._ffi.callback("UwUpdateCallbackType", self._update_callback)
        self._update_callback_func = self._api.uwSetUpdateCallback(self._update_delegate)

        self._shooting_delegate = self._ffi.callback("UwShootingCallbackType", self._shooting_callback)
        self._shooting_callback_func = self._api.uwSetShootingCallback(self._shooting_delegate)

        self._tick = 0

        print(self._api)

        self.prototypes = Prototypes(self._api, self._ffi, self)
        self.map = Map(self._api, self._ffi, self)
        self.world = World(self._api, self._ffi, self)
        self.commands = Commands(self._api, self._ffi)

    def __del__(self):
        self._api.uwDeinitialize()

    def log(self, message: str, severity: Severity = Severity.Info):
        self._api.uwLog(severity.value, _c_str(message))

    def set_player_name(self, name: str):
        self._api.uwSetPlayerName(_c_str(name))

    def set_player_color(self, r: float, g: float, b: float):
        self._api.uwSetPlayerColor(r, g, b)

    def set_start_gui(self, start_gui: bool, extra_params: str = ""):
        self._api.uwSetConnectStartGui(start_gui, _c_str(extra_params))

    def set_connect_find_lan(self, timeout_ms: int = 1000000):
        self._api.uwConnectFindLan(timeout_ms)

    def connect_direct(self, address: str, port: int):
        self._api.uwConnectDirect(_c_str(address), port)

    def connect_lobby_id(self, lobby_id: int):
        self._api.uwConnectLobbyId(lobby_id)

    def connect_new_server(self, visibility: int = 0, name: str = "", extra_params=""):
        self._api.uwConnectNewServer(visibility, _c_str(name), _c_str(extra_params))

    def try_reconnect(self):
        self._api.uwTryReconnect()

    def disconnect(self):
        self._api.uwDisconnect()

    def connection_state_enum(self) -> ConnectionState:
        return ConnectionState(self._api.uwConnectionState())

    def map_state_enum(self) -> MapState:
        return MapState(self._api.uwMapState())

    def tick(self) -> int:
        return self._tick

    def _exception_callback(self, message):
        print(f"Exception: {_to_str(self._ffi, message)}")

    def _log_callback(self, data):
        log_data = LogCallback.from_c(self._ffi, data)
        print(f"{log_data.component}\t{log_data.severity}\t{log_data.message}")

    def add_connection_state_callback(self, callback: Callable[[ConnectionState], None]):
        self._connection_state_changed_handler = callback

    def _connection_state_callback(self, state):
        connection_state = ConnectionState(state)
        print(f"Connection state: {connection_state}")
        for eh in self._connection_state_changed_handler:
            eh(connection_state)

    def add_game_state_callback(self, callback: Callable[[GameState], None]):
        self._game_state_changed_handler.append(callback)

    def _game_state_callback(self, state):
        game_state = GameState(state)
        print(f"Game state: {game_state}")
        for eh in self._game_state_changed_handler:
            eh(game_state)

    def add_map_state_callback(self, callback: Callable[[MapState], None]):
        print(self._map_state_changed_handler)
        self._map_state_changed_handler.append(callback)

    def _map_state_callback(self, state):
        map_state = MapState(state)
        print(f"Map state: {map_state}")
        for eh in self._map_state_changed_handler:
            eh(map_state)

    def add_update_callback(self, callback: Callable[[bool], None]):
        self._updating_handler.append(callback)

    def _update_callback(self, tick: int, stepping: bool):
        self._tick = tick
        for eh in self._updating_handler:
            eh(stepping)

    def add_shooting_callback(self, callback: Callable[[list[ShootingData]], None]):
        self._shooting_handler.append(callback)

    def _shooting_callback(self, shoot_data):
        shooting_data = [ShootingData.from_c(i) for i in _unpack_list(self._ffi, shoot_data, "data")]
        for eh in self._shooting_handler:
            eh(shooting_data)



# coding: utf-8

"""
    Unnatural Worlds API
"""

from setuptools import setup, find_packages  # noqa: H301

# To install the library, run the following
#
# python setup.py install
#
# prerequisite: setuptools
# http://pypi.python.org/pypi/setuptools
NAME = "unnatural-worlds-api"
VERSION = "21.6.2"
PYTHON_REQUIRES = ">=3.7"
REQUIRES = [
]

setup(
    name=NAME,
    version=VERSION,
    description="Unnatural Worlds Bot API",
    author="tivvit",
    author_email="tivvitmail@gmail.com",
    url="",
    keywords=["uw", "unnatural worlds"],
    install_requires=REQUIRES,
    packages=find_packages(exclude=["test", "tests"]),
    include_package_data=True,
    long_description_content_type='text/markdown',
    long_description="""\
    Unnatural Worlds Bot API
    """,  # noqa: E501
    package_data={"": ["bots.h"]},
)
